@@ -2,10 +2,10 @@ import os
 import sys
 import time
 import threading
-import traceback
 import logging
 import itertools
 import random
+import weakref
 
 from collections import deque
 
@@ -16,7 +16,7 @@ import newrelic.core.database_node
 import newrelic.core.error_node
 
 from newrelic.core.stats_engine import CustomMetrics, SampledDataSet
-from newrelic.core.transaction_cache import transaction_cache
+from newrelic.core.trace_cache import trace_cache
 from newrelic.core.thread_utilization import utilization_tracker
 
 from newrelic.core.attribute import (create_attributes,
@@ -43,8 +43,56 @@ DISTRIBUTED_TRACE_TRANSPORT_TYPES = set((
 
 
 class Sentinel(TimeTrace):
-    def __init__(self):
+    def __init__(self, transaction):
         super(Sentinel, self).__init__(None)
+        self.transaction = transaction
+
+        # Set the thread id to the same as the transaction before
+        # saving in the cache.
+        self.thread_id = transaction.thread_id
+        trace_cache().save_trace(self)
+
+    def process_child(self, node, ignore_exclusive=False):
+        if ignore_exclusive:
+            self.children.append(node)
+        else:
+            return super(Sentinel, self).process_child(node)
+
+    def drop_trace(self):
+        trace_cache().drop_trace(self)
+
+    @property
+    def transaction(self):
+        return self._transaction and self._transaction()
+
+    @transaction.setter
+    def transaction(self, value):
+        if value:
+            self._transaction = weakref.ref(value)
+
+    @property
+    def root(self):
+        return self
+
+    @root.setter
+    def root(self, value):
+        pass
+
+
+class CachedPath(object):
+    def __init__(self, transaction):
+        self._name = None
+        self.transaction = weakref.ref(transaction)
+
+    def path(self):
+        if self._name is not None:
+            return self._name
+
+        transaction = self.transaction()
+        if transaction:
+            return transaction.path
+
+        return 'Unknown'
 
 
 class Transaction(object):
@@ -57,7 +105,7 @@ class Transaction(object):
 
         self._application = application
 
-        self.thread_id = transaction_cache().current_thread_id()
+        self.thread_id = None
 
         self._transaction_id = id(self)
         self._transaction_lock = threading.Lock()
@@ -70,12 +118,14 @@ class Transaction(object):
         self._name_priority = 0
         self._group = None
         self._name = None
+        self._cached_path = CachedPath(self)
+        self._loop_time = 0.0
 
         self._frameworks = set()
 
         self._frozen_path = None
 
-        self.current_node = None
+        self.root_span = None
 
         self._request_uri = None
         self._port = None
@@ -86,7 +136,7 @@ class Transaction(object):
         self.end_time = 0.0
         self.last_byte_time = 0.0
 
-        self.total_time = None
+        self.total_time = 0.0
 
         self.stopped = False
 
@@ -132,10 +182,10 @@ class Transaction(object):
         self._calls_write = 0
         self._calls_yield = 0
 
-        self._request_environment = {}
-        self._response_properties = {}
         self._transaction_metrics = {}
 
+        self._agent_attributes = {}
+
         self.background_task = False
 
         self.enabled = False
@@ -147,8 +197,6 @@ class Transaction(object):
 
         self.capture_params = None
 
-        self.response_code = 0
-
         self.apdex = 0
 
         self.rum_token = None
@@ -215,12 +263,6 @@ class Transaction(object):
         if self._state == self.STATE_RUNNING:
             self.__exit__(None, None, None)
 
-    def save_transaction(self):
-        transaction_cache().save_transaction(self)
-
-    def drop_transaction(self):
-        transaction_cache().drop_transaction(self)
-
     def __enter__(self):
 
         assert(self._state == self.STATE_PENDING)
@@ -230,22 +272,6 @@ class Transaction(object):
         if not self.enabled:
             return self
 
-        # Mark transaction as active and update state
-        # used to validate correct usage of class.
-
-        self._state = self.STATE_RUNNING
-
-        # Cache transaction in thread/coroutine local
-        # storage so that it can be accessed from
-        # anywhere in the context of the transaction.
-
-        try:
-            self.save_transaction()
-        except:  # Catch all
-            self._state = self.STATE_PENDING
-            self.enabled = False
-            raise
-
         # Record the start time for transaction.
 
         self.start_time = time.time()
@@ -268,18 +294,20 @@ class Transaction(object):
                 self._thread_utilization_start = \
                         self._utilization_tracker.utilization_count()
 
-        # We need to push an object onto the top of the
-        # node stack so that children can reach back and
-        # add themselves as children to the parent. We
-        # can't use ourself though as we then end up
-        # with a reference count cycle which will cause
-        # the destructor to never be called if the
-        # __exit__() function is never called. We
-        # instead push on to the top of the node stack a
-        # dummy time trace object and when done we will
-        # just grab what we need from that.
+        # Set the thread ID upon entering the transaction.
+        # This is done here so that any asyncio tasks will
+        # be active and the task ID will be used to
+        # store traces into the trace cache.
+        self.thread_id = trace_cache().current_thread_id()
 
-        self.current_node = Sentinel()
+        # Create the root span which pushes itself
+        # into the trace cache as the active trace.
+        self.root_span = Sentinel(self)
+
+        # Mark transaction as active and update state
+        # used to validate correct usage of class.
+
+        self._state = self.STATE_RUNNING
 
         return self
 
@@ -296,42 +324,14 @@ class Transaction(object):
         if not self._settings:
             return
 
-        # Ensure that we are actually back at the top of
-        # transaction call stack. Assume that it is an
-        # instrumentation error and return with hope that
-        # will recover later.
-
-        for _ in range(self._settings.agent_limits.max_outstanding_traces):
-            if isinstance(self.current_node, Sentinel):
-                break
-            self.current_node._force_exit(None, None, None)
-        else:
-            _logger.error('Transaction ended but current_node is not Sentinel.'
-                    ' Current node is %r. Report this issue to New Relic '
-                    'support.\n%s', self.current_node, ''.join(
-                    traceback.format_stack()[:-1]))
-            return
-
-        # Mark as stopped and drop the transaction from
-        # thread/coroutine local storage.
-        #
-        # Note that we validate the saved transaction ID
-        # against that for the current transaction object
-        # to protect against situations where a copy was
-        # made of the transaction object for some reason.
-        # Such a copy when garbage collected could trigger
-        # this function and cause a deadlock if it occurs
-        # while original transaction was being recorded.
+        # Force the root span out of the cache if it's there
+        # This also prevents saving of the root span in the future since the
+        # transaction will be None
+        root = self.root_span
+        root.drop_trace()
 
         self._state = self.STATE_STOPPED
 
-        if not self._dead:
-            try:
-                self.drop_transaction()
-            except:  # Catch all
-                _logger.exception('Unable to drop transaction.')
-                raise
-
         # Record error if one was registered.
 
         if exc is not None and value is not None and tb is not None:
@@ -390,19 +390,13 @@ class Transaction(object):
         # as negative number. Add our own duration to get
         # our own exclusive time.
 
-        root = self.current_node
         children = root.children
 
         exclusive = duration + root.exclusive
 
-        # Calculate total time.
+        # Add transaction exclusive time to total exclusive time
         #
-        # Because we do not track activity on threads, and we currently
-        # don't allocate waiting time in the IOLoop to separate segments
-        # (like External or Datastore), for right now, our total_time is
-        # equal to the duration of the transaction.
-
-        self.total_time = duration
+        self.total_time += exclusive
 
         # Construct final root node of transaction trace.
         # Freeze path in case not already done. This will
@@ -410,22 +404,6 @@ class Transaction(object):
 
         self._freeze_path()
 
-        if self.background_task:
-            transaction_type = 'OtherTransaction'
-        else:
-            transaction_type = 'WebTransaction'
-
-        group = self._group
-
-        if group is None:
-            if self.background_task:
-                group = 'Python'
-            else:
-                group = 'Uri'
-
-        if self.response_code != 0:
-            self._response_properties['STATUS'] = str(self.response_code)
-
         # _sent_end should already be set by this point, but in case it
         # isn't, set it now before we record the custom metrics.
 
@@ -433,27 +411,6 @@ class Transaction(object):
             if not self._sent_end:
                 self._sent_end = time.time()
 
-        if not self.background_task:
-            self.record_custom_metric('Python/WSGI/Input/Bytes',
-                               self._bytes_read)
-            self.record_custom_metric('Python/WSGI/Input/Time',
-                               self.read_duration)
-            self.record_custom_metric('Python/WSGI/Input/Calls/read',
-                               self._calls_read)
-            self.record_custom_metric('Python/WSGI/Input/Calls/readline',
-                               self._calls_readline)
-            self.record_custom_metric('Python/WSGI/Input/Calls/readlines',
-                               self._calls_readlines)
-
-            self.record_custom_metric('Python/WSGI/Output/Bytes',
-                               self._bytes_sent)
-            self.record_custom_metric('Python/WSGI/Output/Time',
-                               self.sent_duration)
-            self.record_custom_metric('Python/WSGI/Output/Calls/yield',
-                               self._calls_yield)
-            self.record_custom_metric('Python/WSGI/Output/Calls/write',
-                               self._calls_write)
-
         if self.client_cross_process_id is not None:
             metric_name = 'ClientApplication/%s/all' % (
                     self.client_cross_process_id)
@@ -474,16 +431,16 @@ class Transaction(object):
             # transaction when distributed tracing or span events are enabled.
             self._compute_sampled_and_priority()
 
+        self._cached_path._name = self.path
         node = newrelic.core.transaction_node.TransactionNode(
                 settings=self._settings,
                 path=self.path,
-                type=transaction_type,
-                group=group,
+                type=self.type,
+                group=self.group_for_metric,
                 base_name=self._name,
                 name_for_metric=self.name_for_metric,
                 port=self._port,
                 request_uri=self._request_uri,
-                response_code=self.response_code,
                 queue_start=self.queue_start,
                 start_time=self.start_time,
                 end_time=self.end_time,
@@ -529,6 +486,7 @@ class Transaction(object):
                 parent_transport_type=self.parent_transport_type,
                 root_span_guid=root.guid,
                 trace_id=self.trace_id,
+                loop_time=self._loop_time,
         )
 
         # Clear settings as we are all done and don't need it
@@ -594,14 +552,7 @@ class Transaction(object):
     @property
     def name_for_metric(self):
         """Combine group and name for use as transaction name in metrics."""
-
-        group = self._group
-
-        if group is None:
-            if self.background_task:
-                group = 'Python'
-            else:
-                group = 'Uri'
+        group = self.group_for_metric
 
         transaction_name = self._name
 
@@ -622,6 +573,18 @@ class Transaction(object):
 
         return name
 
+    @property
+    def group_for_metric(self):
+        _group = self._group
+
+        if _group is None:
+            if self.background_task:
+                _group = 'Python'
+            else:
+                _group = 'Uri'
+
+        return _group
+
     @property
     def path(self):
         if self._frozen_path:
@@ -744,6 +707,8 @@ class Transaction(object):
             i_attrs['synthetics_monitor_id'] = self.synthetics_monitor_id
         if self.total_time:
             i_attrs['totalTime'] = self.total_time
+        if self._loop_time:
+            i_attrs['eventLoopTime'] = self._loop_time
 
         # Add in special CPU time value for UI to display CPU burn.
 
@@ -847,63 +812,16 @@ class Transaction(object):
 
         return attributes_request
 
+    def _add_agent_attribute(self, key, value):
+        self._agent_attributes[key] = value
+
     @property
     def agent_attributes(self):
-        a_attrs = {}
-        settings = self._settings
-        req_env = self._request_environment
-
-        if req_env.get('HTTP_ACCEPT', None):
-            a_attrs['request.headers.accept'] = req_env['HTTP_ACCEPT']
-        if req_env.get('CONTENT_LENGTH', None):
-            a_attrs['request.headers.contentLength'] = \
-                    req_env['CONTENT_LENGTH']
-        if req_env.get('CONTENT_TYPE', None):
-            a_attrs['request.headers.contentType'] = req_env['CONTENT_TYPE']
-        if req_env.get('HTTP_HOST', None):
-            a_attrs['request.headers.host'] = req_env['HTTP_HOST']
-        if req_env.get('HTTP_REFERER', None):
-            a_attrs['request.headers.referer'] = req_env['HTTP_REFERER']
-        if req_env.get('HTTP_USER_AGENT', None):
-            a_attrs['request.headers.userAgent'] = req_env['HTTP_USER_AGENT']
-        if req_env.get('REQUEST_METHOD', None):
-            a_attrs['request.method'] = req_env['REQUEST_METHOD']
-        if self._request_uri:
-            a_attrs['request.uri'] = self._request_uri
-
-        resp_props = self._response_properties
-
-        if resp_props.get('CONTENT_LENGTH', None):
-            a_attrs['response.headers.contentLength'] = \
-                    resp_props['CONTENT_LENGTH']
-        if resp_props.get('CONTENT_TYPE', None):
-            a_attrs['response.headers.contentType'] = \
-                    resp_props['CONTENT_TYPE']
-        if resp_props.get('STATUS', None):
-            a_attrs['response.status'] = resp_props['STATUS']
-
-        if self.read_duration != 0:
-            a_attrs['wsgi.input.seconds'] = self.read_duration
-        if self._bytes_read != 0:
-            a_attrs['wsgi.input.bytes'] = self._bytes_read
-        if self._calls_read != 0:
-            a_attrs['wsgi.input.calls.read'] = self._calls_read
-        if self._calls_readline != 0:
-            a_attrs['wsgi.input.calls.readline'] = self._calls_readline
-        if self._calls_readlines != 0:
-            a_attrs['wsgi.input.calls.readlines'] = self._calls_readlines
-
-        if self.sent_duration != 0:
-            a_attrs['wsgi.output.seconds'] = self.sent_duration
-        if self._bytes_sent != 0:
-            a_attrs['wsgi.output.bytes'] = self._bytes_sent
-        if self._calls_write != 0:
-            a_attrs['wsgi.output.calls.write'] = self._calls_write
-        if self._calls_yield != 0:
-            a_attrs['wsgi.output.calls.yield'] = self._calls_yield
+        a_attrs = self._agent_attributes
 
         if self._settings.process_host.display_name:
-            a_attrs['host.displayName'] = settings.process_host.display_name
+            a_attrs['host.displayName'] = \
+                    self._settings.process_host.display_name
         if self._thread_utilization_value:
             a_attrs['thread.concurrency'] = self._thread_utilization_value
         if self.queue_wait != 0:
@@ -954,7 +872,7 @@ class Transaction(object):
             self._priority = float('%.5f' % random.random())
 
         if self._sampled is None:
-            self._sampled = self._application.compute_sampled(self.priority)
+            self._sampled = self._application.compute_sampled()
             if self._sampled:
                 self._priority += 1
 
@@ -1005,7 +923,13 @@ class Transaction(object):
         m = self._transaction_metrics.get(metric_name, 0)
         self._transaction_metrics[metric_name] = m + 1
 
-    def create_distributed_tracing_payload(self):
+    def _create_distributed_trace_payload_with_guid(self, guid):
+        payload = self.create_distributed_trace_payload()
+        if guid and payload and 'id' in payload['d']:
+            payload['d']['id'] = guid
+        return payload
+
+    def create_distributed_trace_payload(self):
         if not self.enabled:
             return
 
@@ -1016,6 +940,7 @@ class Transaction(object):
 
         if not (account_id and
                 application_id and
+                trusted_account_key and
                 settings.distributed_tracing.enabled):
             return
 
@@ -1035,9 +960,11 @@ class Transaction(object):
             if account_id != trusted_account_key:
                 data['tk'] = trusted_account_key
 
+            current_span = trace_cache().current_trace()
             if (settings.span_events.enabled and
-                    self.current_node and self.sampled):
-                data['id'] = self.current_node.guid
+                    settings.collect_span_events and
+                    current_span and self.sampled):
+                data['id'] = current_span.guid
 
             self.is_distributed_trace = True
 
@@ -1132,7 +1059,14 @@ class Transaction(object):
                 transport_type = 'Unknown'
 
             self.parent_transport_type = transport_type
-            self.parent_transport_duration = time.time() - transport_start
+
+            # If starting in the future, transport duration should be set to 0
+            now = time.time()
+            if transport_start > now:
+                self.parent_transport_duration = 0.0
+            else:
+                self.parent_transport_duration = now - transport_start
+
             self._trace_id = data.get('tr')
 
             if 'pr' in data:
@@ -1212,7 +1146,7 @@ class Transaction(object):
         except Exception:
             pass
 
-    def _generate_response_headers(self):
+    def _generate_response_headers(self, read_length=None):
         nr_headers = []
 
         # Generate metrics and response headers for inbound cross
@@ -1243,8 +1177,13 @@ class Transaction(object):
 
             self._freeze_path()
 
+            if read_length is None:
+                read_length = self._read_length
+
+            read_length = read_length if read_length is not None else -1
+
             payload = (self._settings.cross_process_id, self.path, queue_time,
-                    duration, self._read_length, self.guid, self.record_tt)
+                    duration, read_length, self.guid, self.record_tt)
             app_data = json_encode(payload)
 
             nr_headers.append(('X-NewRelic-App-Data', obfuscate(
@@ -1503,24 +1442,13 @@ class Transaction(object):
         if event:
             self._custom_events.add(event, priority=self.priority)
 
-    def active_node(self):
-        return self.current_node
-
     def _intern_string(self, value):
         return self._string_cache.setdefault(value, value)
 
-    def _push_current(self, node):
-        self.current_node = node
-
-    def _pop_current(self, node):
-        parent = node.parent
-        self.current_node = parent
-
-        return parent
-
     def _process_node(self, node):
         self._trace_node_count += 1
         node.node_count = self._trace_node_count
+        self.total_time += node.exclusive
 
         if type(node) is newrelic.core.database_node.DatabaseNode:
             settings = self._settings
@@ -1578,9 +1506,13 @@ class Transaction(object):
             return True
 
     def add_custom_parameters(self, items):
+        result = True
+
         # items is a list of (name, value) tuples.
         for name, value in items:
-            self.add_custom_parameter(name, value)
+            result &= self.add_custom_parameter(name, value)
+
+        return result
 
     def add_framework_info(self, name, version=None):
         if name:
@@ -1621,12 +1553,10 @@ class Transaction(object):
                 self.autorum_disabled)
         print >> file, 'Supress Apdex: %s' % (
                 self.suppress_apdex)
-        print >> file, 'Current Node: %s' % (
-                self.current_node)
 
 
 def current_transaction(active_only=True):
-    current = transaction_cache().current_transaction()
+    current = trace_cache().current_transaction()
     if active_only:
         if current and (current.ignore_transaction or current.stopped):
             return None
@@ -1680,6 +1610,14 @@ def add_custom_parameter(key, value):
         return False
 
 
+def add_custom_parameters(items):
+    transaction = current_transaction()
+    if transaction:
+        return transaction.add_custom_parameters(items)
+    else:
+        return False
+
+
 def add_framework_info(name, version=None):
     transaction = current_transaction()
     if transaction:
@@ -1730,9 +1668,14 @@ def record_custom_metric(name, value, application=None):
         transaction = current_transaction()
         if transaction:
             transaction.record_custom_metric(name, value)
-    else:
-        if application.enabled:
-            application.record_custom_metric(name, value)
+        else:
+            _logger.debug('record_custom_metric has been called but no '
+                'transaction was running. As a result, the following metric '
+                'has not been recorded. Name: %r Value: %r. To correct this '
+                'problem, supply an application object as a parameter to this '
+                'record_custom_metrics call.', name, value)
+    elif application.enabled:
+        application.record_custom_metric(name, value)
 
 
 def record_custom_metrics(metrics, application=None):
@@ -1740,9 +1683,14 @@ def record_custom_metrics(metrics, application=None):
         transaction = current_transaction()
         if transaction:
             transaction.record_custom_metrics(metrics)
-    else:
-        if application.enabled:
-            application.record_custom_metrics(metrics)
+        else:
+            _logger.debug('record_custom_metrics has been called but no '
+                'transaction was running. As a result, the following metrics '
+                'have not been recorded: %r. To correct this problem, '
+                'supply an application object as a parameter to this '
+                'record_custom_metric call.', list(metrics))
+    elif application.enabled:
+        application.record_custom_metrics(metrics)
 
 
 def record_custom_event(event_type, params, application=None):
@@ -1759,6 +1707,37 @@ def record_custom_event(event_type, params, application=None):
         transaction = current_transaction()
         if transaction:
             transaction.record_custom_event(event_type, params)
-    else:
-        if application.enabled:
-            application.record_custom_event(event_type, params)
+        else:
+            _logger.debug('record_custom_event has been called but no '
+                'transaction was running. As a result, the following event '
+                'has not been recorded. event_type: %r params: %r. To correct '
+                'this problem, supply an application object as a parameter to '
+                'this record_custom_event call.', event_type, params)
+    elif application.enabled:
+        application.record_custom_event(event_type, params)
+
+
+def accept_distributed_trace_payload(payload, transport_type='HTTP'):
+    transaction = current_transaction()
+    if transaction:
+        return transaction.accept_distributed_trace_payload(payload,
+                transport_type)
+    return False
+
+
+def create_distributed_trace_payload():
+    transaction = current_transaction()
+    if transaction:
+        return transaction.create_distributed_trace_payload()
+
+
+def current_trace_id():
+    transaction = current_transaction()
+    if transaction:
+        return transaction.trace_id
+
+
+def current_span_id():
+    trace = trace_cache().current_trace()
+    if trace:
+        return trace.guid
